@@ -15,10 +15,18 @@ class EventTimesSampler(object):
         """
 
         self.d_times = d_times
-        self.times = range(1, self.d_times + 1)
+        self.times = range(1, self.d_times + 2)  # d + 1 for administrative censoring
         self.j_event_types = j_event_types
         self.events = range(1, self.j_event_types + 1)
 
+    def _validate_prob_dfs_list(self, dfs_list: list, numerical_error_tolerance: float = 0.001) -> list:
+        for df in dfs_list:
+            if (((df < (0-numerical_error_tolerance)).any().any()) or ((df > (1+numerical_error_tolerance)).any().any())):
+                raise ValueError("The chosen sampling parameters result in invalid probabilities for event j at time t")
+            # Only fixes numerical errors smaller than the tolerance size
+            df.clip(0, 1, inplace=True)
+        return dfs_list
+
     def calculate_hazards(self, observations_df: pd.DataFrame, hazard_coefs: dict, events: list = None) -> list:
         """
         Calculates the hazard function for the observations given the hazard coefficients.
@@ -42,58 +50,66 @@ class EventTimesSampler(object):
                                   axis=1, keys=(range(1, self.d_times + 1))) for j in events]
         return hazards_dfs
 
-    def calculate_overall_survival(self, hazards: list) -> pd.DataFrame:
+    def calculate_overall_survival(self, hazards: list, numerical_error_tolerance: float = 0.001) -> pd.DataFrame:
         """
         Calculates the overall survival function given the hazards
         Args:
             hazards (list): A list of hazards dataframes for each event type (as returned from EventTimesSampler.calculate_hazards function)
-
+            numerical_error_tolerance (float): Tolerate numerical errors of probabilities up to this value.
         Returns:
             overall_survival (pd.Dataframe): The overall survival functions
         """
+        if (((sum(hazards)) > (1 + numerical_error_tolerance)).sum().sum() > 0):
+            raise ValueError("The chosen sampling parameters result in negative values of the overall survival function")
+        sum_hazards = sum(hazards).clip(0, 1)
         overall_survival = pd.concat([pd.Series(1, index=hazards[0].index),
-                                      (1 - sum(hazards)).cumprod(axis=1).iloc[:, :-1]], axis=1)
+                                      (1 - sum_hazards).cumprod(axis=1).iloc[:, :-1]], axis=1)
         overall_survival.columns += 1
         return overall_survival
 
-    def calculate_prob_event_at_t(self, hazards: list, overall_survival: pd.DataFrame) -> list:
+    def calculate_prob_event_at_t(self, hazards: list, overall_survival: pd.DataFrame,
+                                  numerical_error_tolerance: float = 0.001) -> list:
         """
         Calculates the probability for event j at time t.
 
         Args:
             hazards (list): A list of hazards dataframes for each event type (as returned from EventTimesSampler.calculate_hazards function)
             overall_survival (pd.Dataframe): The overall survival functions
-
+            numerical_error_tolerance (float): Tolerate numerical errors of probabilities up to this value.
         Returns:
             prob_event_at_t (list): A list of dataframes, one for each event type, with the probability of event occurrance at time t to each of the observations.
         """
         prob_event_at_t = [hazard * overall_survival for hazard in hazards]
+        prob_event_at_t = self._validate_prob_dfs_list(prob_event_at_t, numerical_error_tolerance)
         return prob_event_at_t
 
-    def calculate_prob_event_j(self, prob_j_at_t: list) -> list:
+    def calculate_prob_event_j(self, prob_j_at_t: list, numerical_error_tolerance: float = 0.001) -> list:
         """
         Calculates the total probability for event j.
 
         Args:
             prob_j_at_t (list): A list of dataframes, one for each event type, with the probability of event occurrance at time t to each of the observations.
+            numerical_error_tolerance (float): Tolerate numerical errors of probabilities up to this value.
 
         Returns:
             total_prob_j (list): A list of dataframes, one for each event type, with the total probability of event occurrance to each of the observations.
         """
         total_prob_j = [prob.sum(axis=1) for prob in prob_j_at_t]
+        total_prob_j = self._validate_prob_dfs_list(total_prob_j, numerical_error_tolerance)
         return total_prob_j
 
-    def calc_prob_t_given_j(self, prob_j_at_t, total_prob_j):
+    def calc_prob_t_given_j(self, prob_j_at_t, total_prob_j, numerical_error_tolerance=0.001):
         """
         Calculates the conditional probability for event occurrance at time t given J_i=j
         Args:
             prob_j_at_t (list): A list of dataframes, one for each event type, with the probability of event occurrance at time t to each of the observations.
             total_prob_j (list): A list of dataframes, one for each event type, with the total probability of event occurrance to each of the observations.
-
+            numerical_error_tolerance (float): Tolerate numerical errors of probabilities up to this value.
         Returns:
             conditional_prob (list): A list of dataframes, one for each event type, with the conditional probability of event occurrance at t given event type j to each of the observations.
         """
         conditional_prob = [prob.div(sumj, axis=0) for prob, sumj in zip(prob_j_at_t, total_prob_j)]
+        conditional_prob = self._validate_prob_dfs_list(conditional_prob, numerical_error_tolerance)
         return conditional_prob
 
     def sample_event_times(self, observations_df: pd.DataFrame,
@@ -129,7 +145,7 @@ class EventTimesSampler(object):
         observations_df = pd.concat([observations_df, sampled_jt], axis=1)
         return observations_df
 
-    def sample_jt(self, total_prob_j: list, probs_t_given_j: list) -> pd.DataFrame:
+    def sample_jt(self, total_prob_j: list, probs_t_given_j: list, numerical_error_tolerance: float = 0.001) -> pd.DataFrame:
         """
         Sample event type and event time for each observation
         Args:
@@ -140,8 +156,16 @@ class EventTimesSampler(object):
             sampled_df (pd.DataFrame): A dataframe with sampled event time and event type for each observation.
         """
 
+        total_prob_j = self._validate_prob_dfs_list(total_prob_j, numerical_error_tolerance)
+        probs_t_given_j = self._validate_prob_dfs_list(probs_t_given_j, numerical_error_tolerance)
+
         # Add administrative censoring (no event occured until Tmax) probability as J=0
         temp_sums = pd.concat([1 - sum(total_prob_j), *total_prob_j], axis=1, keys=[0, *self.events])
+        if (((temp_sums < (0 - numerical_error_tolerance)).any().any()) or \
+                ((temp_sums > (1 + numerical_error_tolerance)).any().any())):
+            raise ValueError("The chosen sampling parameters result in invalid probabilities")
+        # Only fixes numerical errors smaller than the tolerance size
+        temp_sums.clip(0, 1, inplace=True)
 
         # Efficient way to sample j for each observation with different event probabilities
         sampled_df = (temp_sums.cumsum(1) > np.random.rand(temp_sums.shape[0])[:, None]).idxmax(axis=1).to_frame('J')
@@ -157,8 +181,8 @@ class EventTimesSampler(object):
             # Sample time of occurrence given J_i = j
             temp_ts.append((prob_df.loc[rel_j].cumsum(1) >= np.random.rand(rel_j.shape[0])[:, None]).idxmax(axis=1))
 
-        # Add Tmax for observations with J_i = 0
-        temp_ts.append(pd.Series(self.d_times, index=sampled_df.query('J==0').index))
+        # Add Tmax+1 for observations with J_i = 0
+        temp_ts.append(pd.Series(self.d_times + 1, index=sampled_df.query('J==0').index))
         sampled_df["T"] = pd.concat(temp_ts).sort_index()
         return sampled_df
 
